require(smfsb)
library(matlab)

# StepCLE <- function(N,dt=0.01)
# {
#   S = t(N$Post-N$Pre)
#   v = ncol(S)
#   sdt = sqrt(dt)
#   return(
#     function(x0, t0, deltat,...)
#     {
#       x = x0
#       t = t0
#       termt = t0+deltat
#       repeat {
#         h = N$h(x, t, ...)
#         dw = rnorm(v,0,sdt)
#         dx = S %*% (h*dt + sqrt(h)*dw)
#         x = x + as.vector(dx)
#         if (!any(is.nan(x)))
#           x[x<0] = -x[x<0] # reflection hack
#         t = t+dt
#         if (t > termt)
#           return(x)
#       }
#     }
#   )		
# }

StepCLE <- function(N,dt=0.01)
{
  S = t(N$Post - N$Pre)
  v = ncol(S)
  sdt = sqrt(dt)
  return(function(x0, t0, deltat, ...) {
    x = x0
    t = t0
    termt = t0 + deltat
    repeat {
      h = N$h(x, t, ...)
      dw = rnorm(v, 0, sdt)
      dx = S %*% (h * dt + sqrt(h) * dw)
      x = x + as.vector(dx)
      t = t + dt
      if (t > termt) return(x)
    }
  })
}

options(mc.cores=1)
data(LVdata)

logsumexp <- function(x){
  xmax = which.max(x)
  if (sum(xmax)==0)
  {
    -Inf
  }
  else
  {
    result = log1p(sum(exp(x[-xmax]-x[xmax])))+x[xmax]
    if (is.nan(result))
    {
      -Inf
    }
    else
    {
      result
    }
  }
}

stratified_resample<-function(log_weights)
{
  P = length(log_weights)
  W = exp(log_weights)
  cw = cumsum(W / sum(W))
  n = length(W)
  u = runif(n,0,1)/n
  v = u + (0:(n-1))/n
  
  j = 1
  indices = matrix(0,n)
  for (i in 1:n)
  {
    while(cw[j] < v[i])
    {
      j = j+1
    }
    indices[i] = j
  }
  
  indices
}

stratified_resample_u<-function(log_weights,u)
{
  P = length(log_weights)
  W = exp(log_weights)
  cw = cumsum(W / sum(W))
  n = length(W)
  u = u/n
  v = u + (0:(n-1))/n
  
  j = 1
  indices = matrix(0,n)
  for (i in 1:n)
  {
    while(cw[j] < v[i])
    {
      j = j+1
    }
    indices[i] = j
  }
  
  indices
}

score <- function(epsilon,current_error,alpha,num_imp_u)
{
  sum(current_error<epsilon)-alpha*num_imp_u
}

bisection <- function(range,current_error,alpha,num_imp_u)
{
  current_bisect_epsilon = range[2]
  bisect_size = (range[2] - range[1])/2
  direction = -1
  
  old_score = score(current_bisect_epsilon,current_error,alpha,num_imp_u)
  
  for (i in 1:100)
  {
    new_bisect_epsilon = current_bisect_epsilon + direction*bisect_size
    bisect_size = bisect_size/2
    the_score = score(new_bisect_epsilon,current_error,alpha,num_imp_u)
    direction = -sign(the_score)
    
    current_bisect_epsilon = new_bisect_epsilon
    
    if (the_score==0)
    {
      break
    }
    
  }
  
  current_bisect_epsilon
}

score_u <- function(epsilon,current_error,alpha,num_imp_u,resample_u,current_u)
{
  which_nonzero = which(current_error<epsilon)
  log_weights = matrix(-Inf,num_imp_u)
  log_weights[which_nonzero] = -log(num_imp_u)
  if (all(log_weights==-Inf))
  {
    1-alpha*num_imp_u
  } else {
    indices = stratified_resample_u(log_weights,resample_u)
    current_u = current_u[,indices]
    length(unique(current_u[1,]))-alpha*num_imp_u
  }
}

bisection_u <- function(range,current_error,alpha,num_imp_u,resample_u,current_u)
{
  current_bisect_epsilon = range[2]
  bisect_size = (range[2] - range[1])/2
  direction = -1
  
  old_score = score_u(current_bisect_epsilon,current_error,alpha,num_imp_u,resample_u,current_u)
  
  for (i in 1:100)
  {
    new_bisect_epsilon = current_bisect_epsilon + direction*bisect_size
    bisect_size = bisect_size/2
    the_score = score_u(new_bisect_epsilon,current_error,alpha,num_imp_u,resample_u,current_u)
    direction = -sign(the_score)
    
    current_bisect_epsilon = new_bisect_epsilon
    
    if (the_score==0)
    {
      break
    }
    
  }
  
  current_bisect_epsilon
}

da_score <- function(da_epsilon,proposed_error,old_error,num_kept)
{
  length(which((proposed_error<da_epsilon)&(old_error<da_epsilon)==TRUE))-num_kept
}

da_bisection <- function(range,proposed_error,old_error,num_kept)
{
  current_bisect_epsilon = range[2]
  bisect_size = (range[2] - range[1])/2
  direction = -1
  
  old_score = da_score(current_bisect_epsilon,proposed_error,old_error,num_kept)
  
  for (i in 1:100)
  {
    new_bisect_epsilon = current_bisect_epsilon + direction*bisect_size
    bisect_size = bisect_size/2
    the_score = da_score(new_bisect_epsilon,proposed_error,old_error,num_kept)
    direction = -sign(the_score)
    
    current_bisect_epsilon = new_bisect_epsilon
    
    if (the_score==0)
    {
      break
    }
    
  }
  
  current_bisect_epsilon
}

data(LVdata)
data(spnModels)
N=1e6

#step_size = 0.0005
mystep = StepCLE(LV,dt=step_size)

#reduced_step_size = 0.1
myreducedstep = StepCLE(LV,dt=reduced_step_size)

ssinit <- function(vec)
{
  ac23=as.vector(acf(vec,lag.max=2,plot=FALSE)$acf)[2:3]
  c(mean(vec),log(var(vec)+1),ac23)
}

ssi <- function(ts)
{
  not_nan = !is.nan(ts)
  some_ss = c(ssinit(ts[not_nan[,1],1]),ssinit(ts[not_nan[,2],2]),cor(ts[not_nan[,1],1],ts[not_nan[,2],2]))
  some_ss[is.na(some_ss)] = 0
  some_ss
}

ssinit_new <- function(vec)
{
  n = length(vec)
  ac23=as.vector(acf(vec,lag.max=2,plot=FALSE,type="covariance")$acf[c(2,3)])
  the_var = ((n-1)/n)*var(vec)
  c(mean(vec),the_var,ac23)
}

ssi_new <- function(ts)
{
  first = ssinit_new(ts[,1])
  second = ssinit_new(ts[,2])
  the_n = length(ts[,1])
  c(first,second,(sum( (ts[,1]-first[1])*(ts[,2]-second[1]) )-( sum(ts[,1]-first[1])*sum(ts[,2]-second[1])/the_n ))/the_n )
}

y = LVperfect
ysumi = ssi_new(LVperfect)

sds = c(9.029545e+04,3.744704e+00,2.307471e-01,1.550974e-01,3.554300e+02,1.481975e+00,3.450167e-01,2.652420e-01,6.518170e-01)

ss<-function(ts)
{
  ssi(ts)/sds
}

ss0=ss(LVperfect)

distance <- function(ts)
{
  diff=ss(ts)-ss0
  sum(diff*diff)
}

ss<-function(ts)
{
  ssi(ts)/sds
}

ss_new<-function(ts)
{
  ssi_new(ts)/ysumi
}

ysum = ss0

#num_imp_u = 1000
#num_kept = 100
#alpha = num_kept/num_imp_u
#epsilon = 0.15

prior_sd = 5

total_sims <<- 0
total_sims_trunc <<- 0

current_u = matrix(0,3,num_imp_u)
temp_u = matrix(0,3,num_kept)
temp_u[1,] = exp(runif(num_kept,-6,2))
temp_u[2,] = exp(runif(num_kept,-6,2))
temp_u[3,] = exp(runif(num_kept,-6,2))
current_u = repmat(temp_u,1,num_imp_u/num_kept)

counter = 0

rownames(current_u) = c("th1","th2","th3")
rownames(temp_u) = c("th1","th2","th3")

current_reduced_x = array(0,dim=c(num_imp_u,nrow(LVperfect),ncol(LVperfect)))
temp_reduced_x = array(0,dim=c(num_kept,nrow(LVperfect),ncol(LVperfect)))
for (g in 1:num_kept)
{
  temp_reduced_x[g,,] = simTs(c(x1=50,x2=100),0,30,2,myreducedstep,temp_u[,g])
  total_sims <<- total_sims + 30/reduced_step_size
  total_sims_trunc <<- total_sims_trunc + 2*(length(which(!is.nan(temp_reduced_x[g,,1])))-1)/reduced_step_size
}

current_reduced_x[,,1] = repmat(t(temp_reduced_x[,,1]),num_imp_u/num_kept,1)
current_reduced_x[,,2] = repmat(t(temp_reduced_x[,,2]),num_imp_u/num_kept,1)

current_x = array(0,dim=c(num_imp_u,nrow(LVperfect),ncol(LVperfect)))
temp_x = array(0,dim=c(num_kept,nrow(LVperfect),ncol(LVperfect)))
reduced = TRUE

for (g in 1:num_kept)
{
  temp_x[g,,] = simTs(c(x1=50,x2=100),0,30,2,mystep,temp_u[,g])
  total_sims <<- total_sims + 30/step_size
  total_sims_trunc <<- total_sims_trunc + 2*(length(which(!is.nan(temp_x[g,,1])))-1)/step_size
}

current_x[,,1] = repmat(t(temp_x[,,1]),num_imp_u/num_kept,1)
current_x[,,2] = repmat(t(temp_x[,,2]),num_imp_u/num_kept,1)

current_error = matrix(0,num_imp_u)
for (g in 1:num_imp_u)
{
  current_error[g] = distance(current_x[g,,])
}
current_w = matrix(-log(num_imp_u),num_imp_u)

current_epsilon = Inf

eps_store = c()
unique_before_store = c()
unique_store = c()
total_sims_store = c()
total_sims_trunc_store = c()
num_past_prior_store = c()
num_past_stage1_store = c()
num_accepted_store = c()
da_eps_store = c()
e1_store = c()
e2_store = c()
e3_store = c()
v1_store = c()
v2_store = c()
v3_store = c()

current_epsilon = max(current_error)
eps_store = c(eps_store,current_epsilon)
write.matrix(eps_store,file=paste("epsilon",experiment_number,".txt",sep="_"))
write.matrix(current_u,file=paste("particles",experiment_number,counter,".txt",sep="_"))

while (current_epsilon>epsilon)
{ 
  counter = counter + 1
  current_error[which(is.na(current_error))] = Inf
  
  #browser()
  #current_epsilon = max(current_error)
  old_epsilon = current_epsilon
  
  resample_u = runif(num_imp_u,0,1)
  
  if (length(which(!is.infinite(current_error)))>alpha*num_imp_u)
    current_epsilon = bisection_u(c(0,current_epsilon),current_error,alpha,num_imp_u,resample_u,current_u)#epsilons[e]
  else
    current_epsilon = max(current_error[which(!is.infinite(current_error))])
  
  for (g in 1:num_imp_u)
  {
    if (current_error[g]<current_epsilon)
    {
      current_w[g] = -log(num_imp_u)
    }
    else
    {
      current_w[g] = -Inf
    }
  }
  
  print("Epsilon:")
  print(current_epsilon)
  
  if ( (current_epsilon<epsilon) )
  {
    for (g in 1:num_imp_u)
    {
      if (current_error[g]<epsilon)
      {
        current_w[g] = -log(num_imp_u)
      }
      else
      {
        current_w[g] = -Inf
      }
    }
    
    unique_before_store = c(unique_before_store,length(unique(current_u[1,])))
    print(length(unique(current_u[1,])))
    
    indices = stratified_resample_u(current_w-logsumexp(current_w),resample_u)
    current_x = current_x[indices,,]
    current_reduced_x = current_reduced_x[indices,,]
    current_u = current_u[,indices]
    current_error = current_error[indices]
    current_w = matrix(-log(num_imp_u),num_imp_u)
    
    #hist(current_u[1,])
    #hist(current_u[2,])
    #hist(current_u[3,])
    
    print(total_sims)
    print(total_sims_trunc)
    print(length(unique(current_u[1,])))
    
    prop_cov = cov(t(current_u))
    
    e1_store = c(e1_store,mean(current_u[1,]))
    e2_store = c(e2_store,mean(current_u[2,]))
    e3_store = c(e3_store,mean(current_u[3,]))
    v1_store = c(v1_store,prop_cov[1,1])
    v2_store = c(v2_store,prop_cov[2,2])
    v3_store = c(v3_store,prop_cov[3,3])
    
    eps_store = c(eps_store,epsilon)
    unique_store = c(unique_store,length(unique(current_u[1,])))
    total_sims_store = c(total_sims_store,total_sims)
    total_sims_trunc_store = c(total_sims_trunc_store,total_sims_trunc)
    
    write.matrix(unique_store,file=paste("unique",experiment_number,".txt",sep="_"))
    write.matrix(total_sims_store,file=paste("total_sims",experiment_number,".txt",sep="_"))
    write.matrix(total_sims_trunc_store,file=paste("total_sims_trunc",experiment_number,".txt",sep="_"))
    write.matrix(eps_store,file=paste("epsilon",experiment_number,".txt",sep="_"))
    write.matrix(current_u,file=paste("particles",experiment_number,counter,".txt",sep="_"))
    
    break
  }
  
  unique_before_store = c(unique_before_store,length(unique(current_u[1,])))
  
  indices = stratified_resample_u(current_w-logsumexp(current_w),resample_u)
  current_x[1:num_imp_u,,] = current_x[indices,,]
  current_reduced_x = current_reduced_x[indices,,]
  current_u = current_u[,indices]
  current_error = current_error[indices]
  current_w = matrix(-log(num_imp_u),num_imp_u)
  
  prop_cov = cov(t(current_u))
  
  e1_store = c(e1_store,mean(current_u[1,]))
  e2_store = c(e2_store,mean(current_u[2,]))
  e3_store = c(e3_store,mean(current_u[3,]))
  v1_store = c(v1_store,prop_cov[1,1])
  v2_store = c(v2_store,prop_cov[2,2])
  v3_store = c(v3_store,prop_cov[3,3])
  
  eps_store = c(eps_store,current_epsilon)
  unique_store = c(unique_store,length(unique(current_u[1,])))
  total_sims_store = c(total_sims_store,total_sims)
  total_sims_trunc_store = c(total_sims_trunc_store,total_sims_trunc)
  
  write.matrix(unique_before_store,file=paste("unique_before",experiment_number,".txt",sep="_"))
  write.matrix(unique_store,file=paste("unique",experiment_number,".txt",sep="_"))
  write.matrix(total_sims_store,file=paste("total_sims",experiment_number,".txt",sep="_"))
  write.matrix(total_sims_trunc_store,file=paste("total_sims_trunc",experiment_number,".txt",sep="_"))
  write.matrix(eps_store,file=paste("epsilon",experiment_number,".txt",sep="_"))
  write.matrix(current_u,file=paste("particles",experiment_number,counter,".txt",sep="_"))
  
  number_accepted = 0
  
  all_logu = log(runif(num_imp_u))
  all_proposed_u = matrix(0,3,num_imp_u)
  rownames(all_proposed_u) = c("th1","th2","th3")
  all_logratio = matrix(0,num_imp_u)
  old_reduced_x = array(0,dim=c(num_imp_u,nrow(LVperfect),ncol(LVperfect)))
  old_reduced_error = matrix(Inf,num_imp_u)
  for (g in 1:num_imp_u)
  {
    all_proposed_u[,g] = rtmvnorm(1,mean=current_u[,g],sigma=prop_cov,lower=c(exp(-6),exp(-6),exp(-6)),upper=c(exp(2),exp(2),exp(2)))
    if (all((all_proposed_u[,g]>exp(-6))) && all((all_proposed_u[,g]<exp(2))))
      all_logratio[g] = -sum(log(all_proposed_u[,g])) + dtmvnorm(current_u[,g],mean=all_proposed_u[,g],sigma=prop_cov,lower=c(exp(-6),exp(-6),exp(-6)),upper=c(exp(2),exp(2),exp(2)),log=TRUE) - (-sum(log(current_u[,g])) ) - dtmvnorm(all_proposed_u[,g],mean=current_u[,g],sigma=prop_cov,lower=c(exp(-6),exp(-6),exp(-6)),upper=c(exp(2),exp(2),exp(2)),log=TRUE)
    else
      all_logratio[g] = -Inf
    old_reduced_x[g,,] = current_reduced_x[g,,]
    old_reduced_error[g] = distance(old_reduced_x[g,,])
  }
  
  old_reduced_error[which(is.na(old_reduced_error))] = Inf
  
  which_accept = which(all_logu<all_logratio)
  
  print("Num past prior")
  print(length(which_accept))
  
  num_past_prior_store = c(num_past_prior_store,length(which_accept))
  
  reduced_proposed_x = array(0,dim=c(num_imp_u,nrow(LVperfect),ncol(LVperfect)))
  reduced_proposed_error = matrix(Inf,num_imp_u)

  for (g in 1:length(which_accept))
  {
    reduced_proposed_x[which_accept[g],,] = simTs(c(x1=50,x2=100),0,30,2,myreducedstep,all_proposed_u[,which_accept[g]])
    total_sims <<- total_sims + 30/reduced_step_size
    total_sims_trunc <<- total_sims_trunc + 2*(length(which(!is.nan(reduced_proposed_x[which_accept[g],,1])))-1)/reduced_step_size
    reduced_proposed_error[which_accept[g]] = distance(reduced_proposed_x[which_accept[g],,])
    if (is.na(reduced_proposed_error[which_accept[g]]))
      reduced_proposed_error[which_accept[g]] = Inf
  }

  da_epsilon = max(c(max(reduced_proposed_error[which_accept]),max(old_reduced_error[which_accept])))
  current_score = da_score(da_epsilon,reduced_proposed_error[which_accept],old_reduced_error[which_accept],num_kept)

  if (current_score>0)
  {
    da_epsilon = da_bisection(c(0,da_epsilon),reduced_proposed_error[which_accept],old_reduced_error[which_accept],num_kept)
  }
  
  #if (is.nan(da_epsilon)||is.infinite(da_epsilon))
  #  browser()

  which_accept = which( (reduced_proposed_error<=da_epsilon) & (old_reduced_error<=da_epsilon) )

  da_eps_store = c(da_eps_store,da_epsilon)

  print("DA eps")
  print(da_epsilon)

  print("Num past stage 1")
  print(length(which_accept))

  num_past_stage1_store = c(num_past_stage1_store,length(which_accept))
  
  number_accepted = 0
  
  for (g in 1:length(which_accept))
  {
    proposed_x = simTs(c(x1=50,x2=100),0,30,2,mystep,all_proposed_u[,which_accept[g]])
    total_sims <<- total_sims + 30/step_size
    total_sims_trunc <<- total_sims_trunc + 2*(length(which(!is.nan(proposed_x[,1])))-1)/step_size
    proposed_error = distance(proposed_x)
    if ( (!is.na(proposed_error)) && (proposed_error<current_epsilon) )
    {
      current_u[,which_accept[g]] = all_proposed_u[,which_accept[g]]
      current_x[which_accept[g],,] = proposed_x
      current_reduced_x[which_accept[g],,] = reduced_proposed_x[which_accept[g],,]
      current_error[which_accept[g]] = proposed_error
      number_accepted = number_accepted + 1
    }
  }
  
  print("Number accepted:")
  print(number_accepted)
  
  num_accepted_store = c(num_accepted_store,number_accepted)
  
  write.matrix(da_eps_store,file=paste("da_epsilon",experiment_number,".txt",sep="_"))
  write.matrix(num_past_prior_store,file=paste("num_past_prior",experiment_number,".txt",sep="_"))
  write.matrix(num_past_stage1_store,file=paste("num_past_stage1",experiment_number,".txt",sep="_"))
  write.matrix(num_accepted_store,file=paste("num_accepted",experiment_number,".txt",sep="_"))
  
}
